@@ -7,66 +7,34 @@ import time
 from collections.abc import Generator
 from typing import Any
 
-import async_timeout
 from bleak.backends.device import BLEDevice
+from bleak.exc import BleakError
 
-from .const import IS_LINUX, NO_RSSI_VALUE, RSSI_SWITCH_THRESHOLD
+from .bleak_manager import get_global_bluez_manager_with_timeout
+from .const import (
+    DISCONNECT_TIMEOUT,
+    IS_LINUX,
+    NO_RSSI_VALUE,
+    REAPPEAR_WAIT_INTERVAL,
+    RSSI_SWITCH_THRESHOLD,
+)
+from .util import asyncio_timeout
 
-DISCONNECT_TIMEOUT = 5
-DBUS_CONNECT_TIMEOUT = 8.5
+if IS_LINUX:
+    from dbus_fast.message import Message
 
 _LOGGER = logging.getLogger(__name__)
 
 
-DEFAULT_ATTEMPTS = 2
-
-
 if IS_LINUX:
     with contextlib.suppress(ImportError):  # pragma: no cover
         from bleak.backends.bluezdbus import defs  # pragma: no cover
         from bleak.backends.bluezdbus.manager import (  # pragma: no cover
             BlueZManager,
             DeviceWatcher,
-            get_global_bluez_manager,
         )
 
 
-async def get_global_bluez_manager_with_timeout() -> "BlueZManager" | None:
-    """Get the properties."""
-    if not IS_LINUX:
-        return None
-    if (
-        getattr(get_global_bluez_manager_with_timeout, "_has_dbus_socket", None)
-        is False
-    ):
-        # We are not running on a system with DBus do don't
-        # keep trying to call get_global_bluez_manager as it
-        # waits for a bit trying to connect to DBus.
-        return None
-
-    try:
-        async with async_timeout.timeout(DBUS_CONNECT_TIMEOUT):
-            return await get_global_bluez_manager()
-    except FileNotFoundError as ex:
-        setattr(get_global_bluez_manager_with_timeout, "_has_dbus_socket", False)
-        _LOGGER.debug(
-            "Dbus socket at %s not found, will not try again until next restart: %s",
-            ex.filename,
-            ex,
-        )
-    except asyncio.TimeoutError:
-        setattr(get_global_bluez_manager_with_timeout, "_has_dbus_socket", False)
-        _LOGGER.debug(
-            "Timed out trying to connect to DBus; will not try again until next restart"
-        )
-    except Exception as ex:  # pylint: disable=broad-except
-        _LOGGER.debug(
-            "get_global_bluez_manager_with_timeout failed: %s", ex, exc_info=True
-        )
-
-    return None
-
-
 def device_source(device: BLEDevice) -> str | None:
     """Return the device source."""
     return _device_details_value_or_none(device, "source")
@@ -81,11 +49,6 @@ def _device_details_value_or_none(device: BLEDevice, key: str) -> Any | None:
     return key_value
 
 
-def _reset_dbus_socket_cache() -> None:
-    """Reset the dbus socket cache."""
-    setattr(get_global_bluez_manager_with_timeout, "_has_dbus_socket", None)
-
-
 def adapter_from_path(path: str) -> str:
     """Get the adapter from a ble device path."""
     return path.split("/")[3]
@@ -224,12 +187,90 @@ async def clear_cache(address: str) -> bool:
     """Clear the cache for a device."""
     if not IS_LINUX or not await get_device(address):
         return False
+    caches_cleared: list[str] = []
     with contextlib.suppress(Exception):
-        manager = await get_global_bluez_manager()
-        manager._services_cache.pop(
-            address.upper(), None
-        )  # pylint: disable=protected-access
-        return True
+        if not (manager := await get_global_bluez_manager_with_timeout()):
+            _LOGGER.warning("Failed to clear cache for %s because no manager", address)
+            return False
+        services_cache = manager._services_cache
+        bluez_path = address_to_bluez_path(address)
+        for path in _get_possible_paths(bluez_path):
+            if services_cache.pop(path, None):
+                caches_cleared.append(path)
+        _LOGGER.debug("Cleared cache for %s: %s", address, caches_cleared)
+        async with asyncio_timeout(DISCONNECT_TIMEOUT):
+            for device_path in caches_cleared:
+                # Send since we are going to ignore errors
+                # in case the device is already gone
+                await manager._bus.send(
+                    Message(
+                        destination=defs.BLUEZ_SERVICE,
+                        path=adapter_path_from_device_path(device_path),
+                        interface=defs.ADAPTER_INTERFACE,
+                        member="RemoveDevice",
+                        signature="o",
+                        body=[device_path],
+                    )
+                )
+    return bool(caches_cleared)
+
+
+def adapter_path_from_device_path(device_path: str) -> str:
+    """
+    Scrape the adapter path from a D-Bus device path.
+
+    Args:
+        device_path: The D-Bus object path of the device.
+
+    Returns:
+        A D-Bus object path of the adapter.
+    """
+    # /org/bluez/hci1/dev_FA_23_9D_AA_45_46
+    return device_path[:15]
+
+
+async def wait_for_device_to_reappear(device: BLEDevice, wait_timeout: float) -> bool:
+    """Wait for a device to reappear on the bus."""
+    await asyncio.sleep(0)
+    if (
+        not IS_LINUX
+        or not isinstance(device.details, dict)
+        or "path" not in device.details
+        or not (properties := await _get_properties())
+    ):
+        await asyncio.sleep(wait_timeout)
+        return False
+
+    debug = _LOGGER.isEnabledFor(logging.DEBUG)
+    device_path = address_to_bluez_path(device.address)
+    for i in range(int(wait_timeout / REAPPEAR_WAIT_INTERVAL)):
+        for path in _get_possible_paths(device_path):
+            if path in properties and properties[path].get(defs.DEVICE_INTERFACE):
+                if debug:
+                    _LOGGER.debug(
+                        "%s - %s: Device re-appeared on bus after %s seconds as %s",
+                        device.name,
+                        device.address,
+                        i * REAPPEAR_WAIT_INTERVAL,
+                        path,
+                    )
+                return True
+        if debug:
+            _LOGGER.debug(
+                "%s - %s: Waiting %s/%s for device to re-appear on bus",
+                device.name,
+                device.address,
+                (i + 1) * REAPPEAR_WAIT_INTERVAL,
+                wait_timeout,
+            )
+        await asyncio.sleep(REAPPEAR_WAIT_INTERVAL)
+    if debug:
+        _LOGGER.debug(
+            "%s - %s: Device did not re-appear on bus after %s seconds",
+            device.name,
+            device.address,
+            wait_timeout,
+        )
     return False
 
 
@@ -251,8 +292,14 @@ async def wait_for_disconnect(device: BLEDevice, min_wait_time: float) -> None:
         return
     start = time.monotonic() if min_wait_time else 0
     try:
-        manager = await get_global_bluez_manager()
-        async with async_timeout.timeout(DISCONNECT_TIMEOUT):
+        if not (manager := await get_global_bluez_manager_with_timeout()):
+            _LOGGER.debug(
+                "%s - %s: Failed to wait for disconnect because no manager",
+                device.name,
+                device.address,
+            )
+            return
+        async with asyncio_timeout(DISCONNECT_TIMEOUT):
             await manager._wait_condition(device.details["path"], "Connected", False)
         end = time.monotonic() if min_wait_time else 0
         waited = end - start
@@ -264,7 +311,7 @@ async def wait_for_disconnect(device: BLEDevice, min_wait_time: float) -> None:
         )
         if min_wait_time and waited < min_wait_time:
             await asyncio.sleep(min_wait_time - waited)
-    except KeyError as ex:
+    except (BleakError, KeyError) as ex:
         # Device was removed from bus
         #
         # In testing it was found that most of the CSR adapters
@@ -273,13 +320,14 @@ async def wait_for_disconnect(device: BLEDevice, min_wait_time: float) -> None:
         # When they run out of slots the device they are trying to
         # connect to disappears from the bus so we must backoff
         _LOGGER.debug(
-            "%s - %s: Device was removed from bus, waiting %s for it to re-appear: %s",
+            "%s - %s: Device was removed from bus, waiting %s for it to re-appear: (%s) %s",
             device.name,
             device.address,
             min_wait_time,
+            type(ex),
             ex,
         )
-        await asyncio.sleep(min_wait_time)
+        await wait_for_device_to_reappear(device, min_wait_time)
     except Exception:  # pylint: disable=broad-except
         _LOGGER.debug(
             "%s - %s: Failed waiting for disconnect",
